from collections.abc import Mapping, Sequence
from enum import Enum
from typing import Any, Optional

from dagster._annotations import beta, deprecated
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.assets.definition.asset_spec import AssetSpec
from dagster._core.definitions.metadata.metadata_set import NamespacedMetadataSet, TableMetadataSet
from dagster._record import record
from dagster._utils.cached_method import cached_method
from dagster_shared.serdes import whitelist_for_serdes

from dagster_airbyte.utils import generate_table_schema, get_airbyte_connection_table_name


class AirbyteJobStatusType(str, Enum):
    RUNNING = "running"
    SUCCEEDED = "succeeded"
    CANCELLED = "cancelled"
    PENDING = "pending"
    FAILED = "failed"
    ERROR = "error"
    INCOMPLETE = "incomplete"


@deprecated(breaking_version="1.10", additional_warn_text="Use `AirbyteJobStatusType` instead.")
class AirbyteState:
    RUNNING = AirbyteJobStatusType.RUNNING
    SUCCEEDED = AirbyteJobStatusType.SUCCEEDED
    CANCELLED = AirbyteJobStatusType.CANCELLED
    PENDING = AirbyteJobStatusType.PENDING
    FAILED = AirbyteJobStatusType.FAILED
    ERROR = AirbyteJobStatusType.ERROR
    INCOMPLETE = AirbyteJobStatusType.INCOMPLETE


@record
class AirbyteConnectionTableProps:
    table_name: str
    stream_prefix: Optional[str]
    stream_name: str
    json_schema: Mapping[str, Any]
    connection_id: str
    connection_name: str
    destination_type: Optional[str]
    database: Optional[str]
    schema: Optional[str]

    @property
    def fully_qualified_table_name(self) -> Optional[str]:
        return (
            f"{self.database}.{self.schema}.{self.stream_name}"
            if self.database and self.schema
            else None
        )


@whitelist_for_serdes
@record
class AirbyteConnection:
    """Represents an Airbyte connection, based on data as returned from the API."""

    id: str
    name: str
    stream_prefix: Optional[str]
    streams: Mapping[str, "AirbyteStream"]
    destination_id: str

    @classmethod
    def from_connection_details(
        cls,
        connection_details: Mapping[str, Any],
    ) -> "AirbyteConnection":
        return cls(
            id=connection_details["connectionId"],
            name=connection_details["name"],
            stream_prefix=connection_details.get("prefix"),
            streams={
                stream_details["stream"]["name"]: AirbyteStream.from_stream_details(
                    stream_details=stream_details
                )
                for stream_details in connection_details.get("syncCatalog", {}).get("streams", [])
            },
            destination_id=connection_details["destinationId"],
        )


@whitelist_for_serdes
@record
class AirbyteDestination:
    """Represents an Airbyte destination, based on data as returned from the API."""

    id: str
    type: str
    database: Optional[str]
    schema: Optional[str]

    @classmethod
    def from_destination_details(
        cls,
        destination_details: Mapping[str, Any],
    ) -> "AirbyteDestination":
        return cls(
            id=destination_details["destinationId"],
            type=destination_details["destinationType"],
            database=destination_details["configuration"].get("database"),
            schema=destination_details["configuration"].get("schema"),
        )


@whitelist_for_serdes
@record
class AirbyteStream:
    """Represents an Airbyte stream, based on data as returned from the API.
    A stream in Airbyte corresponds to a table.
    """

    name: str
    selected: bool
    json_schema: Mapping[str, Any]

    @classmethod
    def from_stream_details(
        cls,
        stream_details: Mapping[str, Any],
    ) -> "AirbyteStream":
        return cls(
            name=stream_details["stream"]["name"],
            selected=stream_details["config"].get("selected", False),
            json_schema=stream_details["stream"].get("jsonSchema", {}),
        )


@whitelist_for_serdes
@record
class AirbyteJob:
    """Represents an Airbyte job, based on data as returned from the API."""

    id: int
    status: str

    @classmethod
    def from_job_details(
        cls,
        job_details: Mapping[str, Any],
    ) -> "AirbyteJob":
        return cls(
            id=job_details["jobId"],
            status=job_details["status"],
        )


@whitelist_for_serdes
@record
class AirbyteWorkspaceData:
    """A record representing all content in an Airbyte workspace.
    This applies to both Airbyte OSS and Cloud.
    """

    connections_by_id: Mapping[str, AirbyteConnection]
    destinations_by_id: Mapping[str, AirbyteDestination]

    @cached_method
    def to_airbyte_connection_table_props_data(self) -> Sequence[AirbyteConnectionTableProps]:
        """Method that converts a `AirbyteWorkspaceData` object
        to a collection of `AirbyteConnectionTableProps` objects.
        """
        data: list[AirbyteConnectionTableProps] = []

        for connection in self.connections_by_id.values():
            destination = self.destinations_by_id[connection.destination_id]

            for stream in connection.streams.values():
                if stream.selected:
                    data.append(
                        AirbyteConnectionTableProps(
                            table_name=get_airbyte_connection_table_name(
                                stream_prefix=connection.stream_prefix,
                                stream_name=stream.name,
                            ),
                            stream_prefix=connection.stream_prefix,
                            stream_name=stream.name,
                            json_schema=stream.json_schema,
                            connection_id=connection.id,
                            connection_name=connection.name,
                            destination_type=destination.type,
                            database=destination.database,
                            schema=destination.schema,
                        )
                    )

        return data


class AirbyteMetadataSet(NamespacedMetadataSet):
    connection_id: str
    connection_name: str
    stream_prefix: Optional[str] = None

    @classmethod
    def namespace(cls) -> str:
        return "dagster-airbyte"


@beta
class DagsterAirbyteTranslator:
    """Translator class which converts a `AirbyteConnectionTableProps` object into AssetSpecs.
    Subclass this class to implement custom logic how to translate Airbyte content into asset spec.
    """

    def get_asset_spec(self, props: AirbyteConnectionTableProps) -> AssetSpec:
        """Get the AssetSpec for a table synced by an Airbyte connection."""
        table_schema_props = (
            props.json_schema.get("properties")
            or props.json_schema.get("items", {}).get("properties")
            or {}
        )
        column_schema = generate_table_schema(table_schema_props)

        metadata = {
            **TableMetadataSet(
                column_schema=column_schema,
                table_name=props.fully_qualified_table_name,
            ),
            **AirbyteMetadataSet(
                connection_id=props.connection_id,
                connection_name=props.connection_name,
                stream_prefix=props.stream_prefix,
            ),
        }

        return AssetSpec(
            key=AssetKey(props.table_name),
            metadata=metadata,
            kinds={"airbyte", *({props.destination_type} if props.destination_type else set())},
        )



from collections.abc import Mapping
from typing import Any, NamedTuple, Optional

from dagster._core.definitions.metadata.table import TableSchema


class AirbyteTableMetadata:
    def __init__(
        self,
        raw_table_name: str,
        schema: TableSchema,
        normalization_tables: Optional[Mapping[str, "AirbyteTableMetadata"]] = None,
    ):
        """Contains metadata about an Airbyte table, including its destination raw table name,
        schema and any created normalization tables.
        """
        self.raw_table_name = raw_table_name
        self.schema = schema
        self.normalization_tables = normalization_tables or dict()


class AirbyteOutput(
    NamedTuple(
        "_AirbyteOutput",
        [
            ("job_details", Mapping[str, Any]),
            ("connection_details", Mapping[str, Any]),
        ],
    )
):
    """Contains recorded information about the state of a Airbyte connection job after a sync completes.

    Args:
        job_details (Dict[str, Any]):
            The raw Airbyte API response containing the details of the initiated job. For info
            on the schema of this dictionary, see: https://airbyte-public-api-docs.s3.us-east-2.amazonaws.com/rapidoc-api-docs.html#post-/v1/jobs/get
        connection_details (Dict[str, Any]):
            The raw Airbyte API response containing the details of the sync'd connector. For info
            on the schema of this dictionary, see: https://airbyte-public-api-docs.s3.us-east-2.amazonaws.com/rapidoc-api-docs.html#post-/v1/connections/get
    """



from collections.abc import Sequence
from functools import cached_property
from typing import Annotated, Callable, Optional, Union

import dagster as dg
import pydantic
from dagster._annotations import superseded
from dagster._core.definitions.job_definition import default_job_io_manager
from dagster.components.resolved.base import resolve_fields
from dagster.components.utils.translation import TranslationFn, TranslationFnResolver
from dagster_shared import check

from dagster_airbyte.asset_defs import build_airbyte_assets_definitions
from dagster_airbyte.components.workspace_component.scaffolder import (
    AirbyteWorkspaceComponentScaffolder,
)
from dagster_airbyte.resources import AirbyteCloudWorkspace, AirbyteWorkspace
from dagster_airbyte.translator import (
    AirbyteConnection,
    AirbyteConnectionTableProps,
    DagsterAirbyteTranslator,
)


class ProxyDagsterAirbyteTranslator(DagsterAirbyteTranslator):
    def __init__(self, fn: TranslationFn[AirbyteConnectionTableProps]):
        self.fn = fn

    def get_asset_spec(self, props: AirbyteConnectionTableProps) -> dg.AssetSpec:
        base_asset_spec = super().get_asset_spec(props)
        spec = self.fn(base_asset_spec, props)

        return spec


class BaseAirbyteWorkspaceModel(dg.Model, dg.Resolvable):
    request_max_retries: Annotated[
        int,
        pydantic.Field(
            default=3,
            description=(
                "The maximum number of times requests to the Airbyte API should be retried "
                "before failing."
            ),
        ),
    ]
    request_retry_delay: Annotated[
        float,
        pydantic.Field(
            default=0.25,
            description="Time (in seconds) to wait between each request retry.",
        ),
    ]
    request_timeout: Annotated[
        int,
        pydantic.Field(
            default=15,
            description="Time (in seconds) after which the requests to Airbyte are declared timed out.",
        ),
    ]


class AirbyteWorkspaceModel(BaseAirbyteWorkspaceModel):
    rest_api_base_url: Annotated[
        str,
        pydantic.Field(
            ...,
            description=(
                "The base URL for the Airbyte REST API. "
                "For Airbyte Cloud, leave this as the default. "
                "For self-managed Airbyte, this is usually <your Airbyte host>/api/public/v1."
            ),
            examples=[
                "http://localhost:8000/api/public/v1",
                "https://my-airbyte-server.com/api/public/v1",
                "http://airbyte-airbyte-server-svc.airbyte.svc.cluster.local:8001/api/public/v1",
            ],
        ),
    ]
    configuration_api_base_url: Annotated[
        str,
        pydantic.Field(
            ...,
            description=(
                "The base URL for the Airbyte Configuration API. "
                "For Airbyte Cloud, leave this as the default. "
                "For self-managed Airbyte, this is usually <your Airbyte host>/api/v1."
            ),
            examples=[
                "http://localhost:8000/api/v1",
                "https://my-airbyte-server.com/api/v1",
                "http://airbyte-airbyte-server-svc.airbyte.svc.cluster.local:8001/api/v1",
            ],
        ),
    ]
    workspace_id: Annotated[str, pydantic.Field(..., description="The Airbyte workspace ID.")]
    client_id: Annotated[
        Optional[str],
        pydantic.Field(None, description="Client ID used to authenticate to Airbyte."),
    ]
    client_secret: Annotated[
        Optional[str],
        pydantic.Field(None, description="Client secret used to authenticate to Airbyte."),
    ]
    username: Annotated[
        Optional[str],
        pydantic.Field(
            None,
            description="Username used to authenticate to Airbyte. Used for self-managed Airbyte with basic auth.",
        ),
    ]
    password: Annotated[
        Optional[str],
        pydantic.Field(
            None,
            description="Password used to authenticate to Airbyte. Used for self-managed Airbyte with basic auth.",
        ),
    ]


class AirbyteCloudWorkspaceModel(BaseAirbyteWorkspaceModel):
    workspace_id: Annotated[str, pydantic.Field(..., description="The Airbyte workspace ID.")]
    client_id: Annotated[
        Optional[str],
        pydantic.Field(..., description="Client ID used to authenticate to Airbyte."),
    ]
    client_secret: Annotated[
        Optional[str],
        pydantic.Field(..., description="Client secret used to authenticate to Airbyte."),
    ]


class AirbyteConnectionSelectorByName(dg.Model):
    by_name: Annotated[
        Sequence[str],
        pydantic.Field(..., description="A list of connection names to include in the collection."),
    ]


class AirbyteConnectionSelectorById(dg.Model):
    by_id: Annotated[
        Sequence[str],
        pydantic.Field(..., description="A list of connection IDs to include in the collection."),
    ]


def resolve_connection_selector(
    context: dg.ResolutionContext, model
) -> Optional[Callable[[AirbyteConnection], bool]]:
    if isinstance(model, str):
        model = context.resolve_value(model)

    if isinstance(model, AirbyteConnectionSelectorByName):
        return lambda connection: connection.name in model.by_name
    elif isinstance(model, AirbyteConnectionSelectorById):
        return lambda connection: connection.id in model.by_id
    else:
        check.failed(f"Unknown connection target type: {type(model)}")


def resolve_airbyte_workspace_type(context: dg.ResolutionContext, model):
    if isinstance(model, AirbyteWorkspaceModel):
        return AirbyteWorkspace(**resolve_fields(model, AirbyteWorkspaceModel, context))
    elif isinstance(model, AirbyteCloudWorkspaceModel):
        return AirbyteCloudWorkspace(**resolve_fields(model, AirbyteCloudWorkspaceModel, context))
    else:
        check.failed(f"Unknown Airbyte workspace type: {type(model)}")


@dg.scaffold_with(AirbyteWorkspaceComponentScaffolder)
class AirbyteWorkspaceComponent(dg.Component, dg.Model, dg.Resolvable):
    """Loads Airbyte connections from a given Airbyte workspace as Dagster assets.
    Materializing these assets will trigger a sync of the Airbyte connection, enabling
    you to schedule Airbyte syncs using Dagster.
    """

    workspace: Annotated[
        Union[AirbyteWorkspace, AirbyteCloudWorkspace],
        dg.Resolver(
            resolve_airbyte_workspace_type,
            model_field_type=Union[AirbyteWorkspaceModel, AirbyteCloudWorkspaceModel],
        ),
    ]
    connection_selector: Annotated[
        Optional[Callable[[AirbyteConnection], bool]],
        dg.Resolver(
            resolve_connection_selector,
            model_field_type=Union[
                str, AirbyteConnectionSelectorByName, AirbyteConnectionSelectorById
            ],
            description="Function used to select Airbyte connections to pull into Dagster.",
        ),
    ] = None
    translation: Optional[
        Annotated[
            TranslationFn[AirbyteConnectionTableProps],
            TranslationFnResolver(template_vars_for_translation_fn=lambda data: {"props": data}),
        ]
    ] = pydantic.Field(
        None,
        description="Function used to translate Airbyte connection table properties into Dagster asset specs.",
    )

    @cached_property
    def translator(self) -> DagsterAirbyteTranslator:
        if self.translation:
            return ProxyDagsterAirbyteTranslator(self.translation)
        return DagsterAirbyteTranslator()

    def build_defs(self, context: dg.ComponentLoadContext) -> dg.Definitions:
        airbyte_assets = build_airbyte_assets_definitions(
            workspace=self.workspace,
            dagster_airbyte_translator=self.translator,
            connection_selector_fn=self.connection_selector,
        )
        assets_with_resource = [
            airbyte_asset.with_resources(
                {
                    "airbyte": self.workspace.get_resource_definition(),
                    "io_manager": default_job_io_manager,
                }
            )
            for airbyte_asset in airbyte_assets
        ]
        return dg.Definitions(assets=assets_with_resource)


# Subclassing to create the alias to be able to use the superseded decorator.
@superseded(additional_warn_text="Superseded. Use AirbyteWorkspaceComponent instead.")
class AirbyteCloudWorkspaceComponent(AirbyteWorkspaceComponent): ...
